from unittest.mock import create_autospec
from unittest.mock import patch

import pytest
from packaging.version import Version
from pypi_simple import DistributionPackage
from pypi_simple import NoSuchProjectError
from pypi_simple.client import PyPISimple

from piplexed.pypi.pypi_info import find_most_recent_version_on_pypi
from piplexed.pypi.pypi_info import get_pypi_versions
from piplexed.pypi.pypi_info import latest_pypi_version
from piplexed.pypi.pypi_info import pypi_package_info
from piplexed.venvs.pipx_venvs import PackageInfo


@pytest.mark.parametrize(
    "packages_data, is_prerelease, expected",
    [
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0b0", "sdist", False),
            ],
            False,
            Version("1.0.0"),
            id="no pre-releases",
        ),
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0b0", "sdist", False),
            ],
            True,
            Version("2.0b0"),
            id="pre-release allowed",
        ),
    ],
)
def test_get_latest_version_prerelease_flag(packages_data, is_prerelease, expected):
    packages = [
        create_autospec(DistributionPackage, version=v, package_type=t, is_yanked=y) for v, t, y in packages_data
    ]
    assert latest_pypi_version(packages, is_prerelease=is_prerelease) == expected


@pytest.mark.parametrize(
    "packages_data, is_prerelease, expected",
    [
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", False),
                ("2.5.0", "sdist", False),
            ],
            False,
            Version("2.5.0"),
            id="standard sdists",
        ),
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", False),
                ("2.5.0", "sdist", True),
            ],
            False,
            Version("2.0.0"),
            id="yanked sdists",
        ),
        pytest.param(
            [
                ("2.0.0", "sdist", False),
                ("2.5.0", "sdist", False),
                ("1.0.0", "sdist", False),
            ],
            False,
            Version("2.5.0"),
            id="jumbled sdists",
        ),
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", False),
                ("2.5b0", "sdist", False),
            ],
            False,
            Version("2.0.0"),
            id="pre_lease sdist",
        ),
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", True),
                ("2.5b0", "sdist", False),
            ],
            False,
            Version("1.0.0"),
            id="pre_release and yanked sdists",
        ),
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("invalid", "sdist", False),
            ],
            False,
            Version("1.0.0"),
            id="Invalid version does not fail",
        ),
    ],
)
def test_get_latest_version_sdists(packages_data, is_prerelease, expected):
    """
    Test examples are all sdists as piplexed only finds versions for sdists
    """
    packages = [
        create_autospec(DistributionPackage, version=v, package_type=t, is_yanked=y) for v, t, y in packages_data
    ]
    assert latest_pypi_version(packages, is_prerelease=is_prerelease) == expected


@pytest.mark.parametrize(
    "packages_data, is_prerelease, expected",
    [
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", False),
                ("2.5.0", "wheel", False),
            ],
            False,
            Version("2.0.0"),
            id="single wheel",
        ),
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", False),
                ("2.0.0", "wheel", False),
            ],
            False,
            Version("2.0.0"),
            id="sdist and wheel",
        ),
        pytest.param(
            [
                ("2.0.0", "sdist", False),
                ("2.5.0", "wheel", False),
                ("1.0.0", "sdist", False),
            ],
            False,
            Version("2.0.0"),
            id="jumbled single wheel",
        ),
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", False),
                ("2.5.0", "wheel", True),
            ],
            False,
            Version("2.0.0"),
            id="single yanked wheel",
        ),
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", True),
                ("2.5.0", "wheel", False),
            ],
            False,
            Version("1.0.0"),
            id="single wheel yanked sdist",
        ),
    ],
)
def test_get_latest_version_wheels(packages_data, is_prerelease, expected):
    """
    Piplexed only finds versions for sdists, this test, this test ensures wheels
    aren't included.
    """
    packages = [
        create_autospec(DistributionPackage, version=v, package_type=t, is_yanked=y) for v, t, y in packages_data
    ]
    assert latest_pypi_version(packages, is_prerelease=is_prerelease) == expected


@patch("piplexed.pypi.pypi_info.pypi_package_info")
def test_get_pypi_versions(mock_page):
    pypi_packages_data = [
        ("1.0.0", "sdist", False),
        ("2.0.0", "sdist", False),
        ("2.5.0", "sdist", False),
    ]
    pypi_packages = [
        create_autospec(DistributionPackage, version=v, package_type=t, is_yanked=y) for v, t, y in pypi_packages_data
    ]

    mock_page.return_value = pypi_packages

    mock_client = create_autospec(PyPISimple, spec_set=True)
    package = PackageInfo(name="testproj", version=Version("0.5.0"), python=None, latest_pypi_version=None)

    result = get_pypi_versions(client=mock_client, package=package, is_prerelease=False)
    assert [result] == [
        PackageInfo(name="testproj", version=Version("0.5.0"), python=None, latest_pypi_version=Version("2.5.0"))
    ]


# -----------------------------------------------------
# To test find_most_recent_version_on_pypi need to mock some data


@pytest.fixture
def mocked_venvs():
    return [
        PackageInfo(name="package1", version="1.0.0"),
        PackageInfo(name="package2", version="2.0.0"),
        PackageInfo(name="nonpypi-package", version="1.5.1"),
    ]


@pytest.fixture
def mock_get_pypi_versions():
    """
    In find_most_recent_version_on_pypi get_pypi_version is called for each package
    returned from get_pipx_metadata. Use side effects to get different results
    for each package.
    """
    with patch("piplexed.pypi.pypi_info.get_pypi_versions") as mock:

        def side_effect(client, pkg, is_prerelease):  # noqa: ARG001
            if pkg.name == "package1":
                return PackageInfo(
                    name="package1", version="1.0.0", latest_pypi_version="1.1.0" if not is_prerelease else "1.2.0-beta"
                )
            elif pkg.name == "package2":
                return PackageInfo(name="package2", version="2.0.0", latest_pypi_version="2.0.0")
            else:
                raise NoSuchProjectError(project="nonpypi-package", url="mock/url")

        mock.side_effect = side_effect
        yield mock


def test_find_most_recent_version_on_pypi(mocked_venvs, mock_get_pypi_versions):
    result = find_most_recent_version_on_pypi(venvs=mocked_venvs, is_prerelease=False)
    assert len(result) == 1
    assert result[0].name == "package1"
    assert result[0].version == "1.0.0"
    assert result[0].latest_pypi_version == "1.1.0"
    assert mock_get_pypi_versions.call_count == 3


def test_find_most_recent_version_on_pypi_is_prerelease(mocked_venvs, mock_get_pypi_versions):
    result = find_most_recent_version_on_pypi(venvs=mocked_venvs, is_prerelease=True)
    assert len(result) == 1
    assert result[0].name == "package1"
    assert result[0].version == "1.0.0"
    assert result[0].latest_pypi_version == "1.2.0-beta"
    assert mock_get_pypi_versions.call_count == 3


def test_find_most_recent_version_on_pypi_no_updates(mocked_venvs, mock_get_pypi_versions):
    mock_get_pypi_versions.side_effect = lambda client, pkg, is_prerelease: (  # noqa: ARG005
        PackageInfo(name="package1", version="1.0.0", latest_pypi_version="1.0.0" if is_prerelease else "1.0.0")
        if pkg.name == "package1"
        else PackageInfo(name="package2", version="2.0.0", latest_pypi_version="2.0.0")
    )

    result = find_most_recent_version_on_pypi(venvs=mocked_venvs, is_prerelease=False)
    assert not result
    assert mock_get_pypi_versions.call_count == 3


def test_find_most_recent_version_on_pypi_unstable_no_updates(mocked_venvs, mock_get_pypi_versions):
    mock_get_pypi_versions.side_effect = lambda client, pkg, is_prerelease: (  # noqa: ARG005
        PackageInfo(name="package1", version="1.0.0", latest_pypi_version="1.0.0" if is_prerelease else "1.0.0")
        if pkg.name == "package1"
        else PackageInfo(name="package2", version="2.0.0", latest_pypi_version="2.0.0")
    )

    result = find_most_recent_version_on_pypi(venvs=mocked_venvs, is_prerelease=True)
    assert not result
    assert mock_get_pypi_versions.call_count == 3


def test_pypi_package_info_non_pypi_project():
    mock_client = create_autospec(PyPISimple, instance=True)
    mock_client.get_project_page.side_effect = NoSuchProjectError("jeff", "mock/url")
    with pytest.raises(NoSuchProjectError):
        pypi_package_info(client=mock_client, package_name="jeff")



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Any
from unittest.mock import patch

import pytest
from packaging.version import Version

from piplexed.venvs import PackageInfo
from piplexed.venvs.pipx_venvs import PIPX_METADATA_VERSIONS
from piplexed.venvs.pipx_venvs import get_local_venv
from piplexed.venvs.pipx_venvs import installed_pipx_tools
from piplexed.venvs.pipx_venvs import is_metadata_version_valid
from piplexed.venvs.pipx_venvs import pipx_home_paths_for_os

# Metadata changes for 0.1 -> 0.3 and 0.4 -> 0.5 were at package level
MOCK_BASE_PIPX_METADATA: dict[str, Any] = {
    "main_package": None,
    "python_version": None,
    "venv_args": [],
    "injected_packages": {},
    "pipx_metadata_version": "0.1",
}

MOCK_PIPX_METADATA_0_4: dict[str, Any] = MOCK_BASE_PIPX_METADATA | {"source_interpreter": None}

MOCK_PACKAGE_DATA_0_1: dict[str, Any] = {
    "package": None,
    "package_or_url": None,
    "pip_args": [],
    "include_dependencies": False,
    "include_apps": True,
    "apps": [],
    "app_paths": [],
    "apps_of_dependencies": [],
    "app_paths_of_dependencies": {},
    "package_version": "",
}

MOCK_PACKAGE_DATA_0_2 = MOCK_PACKAGE_DATA_0_1 | {"suffix": ""}

MOCK_PACKAGE_DATA_0_3_and_0_4 = MOCK_PACKAGE_DATA_0_2 | {
    "man_pages": [],
    "man_paths": [],
    "man_pages_of_dependencies": [],
    "man_paths_of_dependencies": {},
}

MOCK_PACKAGE_DATA_0_5 = MOCK_PACKAGE_DATA_0_3_and_0_4 | {"pinned": False}


def mock_metadata(metadata_version: str, pypi_package: bool = True) -> dict[str, Any]:  # noqa: FBT001, FBT002
    if metadata_version in ["0.1", "0.2", "0.3"]:
        metadata_template = MOCK_BASE_PIPX_METADATA
    elif metadata_version in ["0.4", "0.5"]:
        metadata_template = MOCK_PIPX_METADATA_0_4
    else:
        err_msg = f"Internal Test Error: Unknown metadata_version={metadata_version}"
        raise Exception(err_msg)

    if metadata_version == "0.1":
        package_template = MOCK_PACKAGE_DATA_0_1
    elif metadata_version == "0.2":
        package_template = MOCK_PACKAGE_DATA_0_2
    elif metadata_version in ["0.3", "0.4"]:
        package_template = MOCK_PACKAGE_DATA_0_3_and_0_4
    elif metadata_version == "0.5":
        package_template = MOCK_PACKAGE_DATA_0_5
    else:
        err_msg = f"Internal Test Error: Unknown metadata_version={metadata_version}"
        raise Exception(err_msg)

    if pypi_package:
        package_template["package"] = "Testy_McTestFace"
        package_template["package_or_url"] = "Testy_McTestFace"
    else:
        package_template["package"] = "Local_Package"
        package_template["package_or_url"] = "path/to/local/wheel"

    package_template["package_version"] = "23.1.0"
    metadata_template["pipx_metadata_version"] = metadata_version
    metadata_template["python_version"] = "Python 3.11.2"
    metadata_template["main_package"] = package_template

    return metadata_template


@pytest.fixture
def venv_dir_test_setup(tmp_path):
    env_dir = tmp_path / "venvs"
    env_dir.mkdir()
    pypi_package = env_dir / "pypi_package"
    pypi_package.mkdir()
    local_package = env_dir / "local_package"
    local_package.mkdir()

    return env_dir


@pytest.mark.parametrize("pipx_metadata_version", ("0.1", "0.2", "0.3", "0.4", "0.5"))
def test_pipx_metadata(venv_dir_test_setup, pipx_metadata_version):
    pypi_json = venv_dir_test_setup / "pypi_package" / "pipx_metadata.json"
    local_json = venv_dir_test_setup / "local_package" / "pipx_metadata.json"

    expected = [
        PackageInfo(name="local-package", version=Version("23.1.0"), python="3.11.2"),
        PackageInfo(name="testy-mctestface", version=Version("23.1.0"), python="3.11.2"),
    ]

    with open(pypi_json, "w") as f:
        pipx_metadata = mock_metadata(metadata_version=pipx_metadata_version)
        json.dump(pipx_metadata, f)

    with open(local_json, "w") as f:
        pipx_metadata = mock_metadata(metadata_version=pipx_metadata_version, pypi_package=False)
        json.dump(pipx_metadata, f)

    results = installed_pipx_tools(venv_dir_test_setup)
    assert expected[0] in results
    assert expected[1] in results
    assert len(results) == len(expected)


def test_multiple_json_files_in_venv(venv_dir_test_setup):
    pypi_json = venv_dir_test_setup / "pypi_package" / "pipx_metadata.json"
    extra_json = venv_dir_test_setup / "pypi_package" / "schema.json"

    expected = [PackageInfo(name="testy-mctestface", version=Version("23.1.0"), python="3.11.2")]

    with open(pypi_json, "w") as f:
        pipx_metadata = mock_metadata(metadata_version="0.5")
        json.dump(pipx_metadata, f)

    with open(extra_json, "w") as f:
        extra_data = {"test": "test extra data"}
        json.dump(extra_data, f)

    assert installed_pipx_tools(venv_dir_test_setup) == expected


def test_venv_dir_is_none():
    assert installed_pipx_tools(venv_dir=None) == []


def test_venv_dir_not_exists(tmp_path):
    non_existent_path = tmp_path / "joker"
    assert installed_pipx_tools(venv_dir=non_existent_path) == []


@pytest.fixture
def mock_user_data_path():
    with patch("piplexed.venvs.pipx_venvs.user_data_path") as mock_user_data:
        mock_user_data.return_value = "mock/user/data/pipx"
        yield mock_user_data


def test_pipx_home_paths_for_os_linux(mock_user_data_path):  # noqa: ARG001
    default, fallback = pipx_home_paths_for_os("Linux")
    assert str(default) == os.path.join("mock", "user", "data", "pipx")
    assert [str(p) for p in fallback] == [
        os.path.join(str(Path.home()), ".local", "pipx"),
    ]


def test_pipx_home_paths_for_os_windows(mock_user_data_path):  # noqa: ARG001
    default, fallback = pipx_home_paths_for_os("Windows")
    assert str(default) == os.path.join(str(Path.home()), "pipx")
    assert [str(p) for p in fallback] == [
        os.path.join(str(Path.home()), ".local", "pipx"),
        os.path.join("mock", "user", "data", "pipx"),
    ]


def test_pipx_home_paths_for_os_other(mock_user_data_path):  # noqa: ARG001
    default, fallback = pipx_home_paths_for_os("Darwin")
    assert str(default) == os.path.join(str(Path.home()), ".local", "pipx")
    assert [str(p) for p in fallback] == [os.path.join("mock", "user", "data", "pipx")]


@pytest.fixture
def mock_pipx_homes(monkeypatch, tmp_path):
    default_home = tmp_path / "default"
    fallback_homes = [tmp_path / "fallback1", tmp_path / "fallback2"]

    monkeypatch.setattr("piplexed.venvs.pipx_venvs.DEFAULT_PIPX_HOME", default_home)
    monkeypatch.setattr("piplexed.venvs.pipx_venvs.FALLBACK_PIPX_HOMES", fallback_homes)

    return (default_home, fallback_homes)


def test_default_pipx_home_exists(mock_pipx_homes):
    default_home, _ = mock_pipx_homes
    default_home.mkdir()

    result = get_local_venv()
    assert result == default_home / "venvs"


def test_default_pipx_home_not_exists_fallback1_exists(mock_pipx_homes):
    _, fallbacks = mock_pipx_homes
    fallback_1 = fallbacks[0]
    fallback_1.mkdir()
    result = get_local_venv()
    assert result == fallback_1 / "venvs"


def test_default_pipx_home_not_exists_fallback2_exists(mock_pipx_homes):
    _, fallbacks = mock_pipx_homes
    fallback_2 = fallbacks[1]
    fallback_2.mkdir()
    result = get_local_venv()
    assert result == fallback_2 / "venvs"


def test_default_pipx_home_not_exists_fallback1_and_fallback2_exists(mock_pipx_homes):
    _, fallbacks = mock_pipx_homes
    fallback_1 = fallbacks[0]
    fallback_2 = fallbacks[1]
    fallback_1.mkdir()
    fallback_2.mkdir()
    result = get_local_venv()
    assert result == fallback_1 / "venvs"


def test_default_pipx_home_exists_and_fallback_exists(mock_pipx_homes):
    default, fallbacks = mock_pipx_homes
    fallback_1 = fallbacks[0]
    fallback_2 = fallbacks[1]
    default.mkdir()
    fallback_1.mkdir()
    fallback_2.mkdir()
    result = get_local_venv()
    assert result == default / "venvs"


def test_default_pipx_home_not_exists_and_fallback_not_exists(mock_pipx_homes):
    default, fallbacks = mock_pipx_homes
    result = get_local_venv()
    assert result is None


@pytest.mark.parametrize(
    "metadata_version, expected",
    (
        ("0.1", True),
        ("0.2", True),
        ("0.3", True),
        ("0.4", True),
        ("0.5", True),
        ("0.6", False),
        ("", False),
        (None, False),
    ),
)
def test_pipx_metaversion_validation(metadata_version, expected):
    assert is_metadata_version_valid(metadata_version, PIPX_METADATA_VERSIONS) == expected


def test_pipx_metadata_warning(venv_dir_test_setup):
    pypi_json = venv_dir_test_setup / "pypi_package" / "pipx_metadata.json"

    with open(pypi_json, "w") as f:
        pipx_metadata = mock_metadata(metadata_version="0.5")
        pipx_metadata["pipx_metadata_version"] = None
        json.dump(pipx_metadata, f)

    with pytest.warns(UserWarning):
        installed_pipx_tools(venv_dir_test_setup)


def test_newer_pypi_version():
    package = PackageInfo("package1", "1.0.0", None, "1.1.0")
    assert package.newer_pypi_version()


def test_pypi_version_no_change():
    package = PackageInfo("package1", "1.0.0", None, "1.0.0")
    assert not package.newer_pypi_version()


def test_pypi_version_no_value():
    package = PackageInfo("package1", "1.0.0", None, None)
    assert not package.newer_pypi_version()



"""Logic for orchestrating what the cli flags do"""

from piplexed.display.print_table import print_installed_table
from piplexed.display.print_table import print_outdated_table
from piplexed.display.print_tree import print_installed_tree
from piplexed.display.print_tree import print_outdated_tree
from piplexed.pypi.pypi_info import find_most_recent_version_on_pypi
from piplexed.utils import rich_format_error
from piplexed.utils import rich_format_info
from piplexed.venvs import PackageInfo
from piplexed.venvs import ToolType
from piplexed.venvs.pipx_venvs import PIPX_LOCAL_VENVS
from piplexed.venvs.pipx_venvs import installed_pipx_tools
from piplexed.venvs.uv_venvs import HAS_UV
from piplexed.venvs.uv_venvs import UV_TOOL_DIR
from piplexed.venvs.uv_venvs import installed_uv_tools


def find_installed_tools(tool: ToolType) -> list[PackageInfo]:
    if tool == "pipx":
        venvs = installed_pipx_tools()
        if not venvs and PIPX_LOCAL_VENVS is None:
            rich_format_error("Looking for pipx tools.\nUnable to locate pipx venv directory.")
        elif not venvs:
            rich_format_info(f"No installed tools found in {PIPX_LOCAL_VENVS}")
    elif tool == "uv":
        venvs = installed_uv_tools()
        if not venvs and not HAS_UV:
            rich_format_error("Looking for uv tools.\nUnable to find uv on your path.")
        elif not venvs and HAS_UV and not UV_TOOL_DIR:
            rich_format_error(
                "Looking uv tools.\n"
                "Unable to find a uv tool directory.\nHave you installed a package using 'uv tool install [PACKAGE]' ?"
            )
    else:
        pipx_venvs = installed_pipx_tools()
        uv_venvs = installed_uv_tools()
        if not pipx_venvs:
            rich_format_info(
                f"Looking for {tool.value} tools installed with pipx and/or uv.\nUnable to find pipx venvs"
            )

        if not uv_venvs and not HAS_UV:
            rich_format_error(
                f"Looking for {tool.value} tools installed with pipx and/or uv.\nnable to find uv on your path."
            )
        if not uv_venvs and HAS_UV and not UV_TOOL_DIR:
            rich_format_info(
                f"Looking for {tool.value} tools installed with pipx and/or uv.\nUnable to find a uv tool directory."
            )

        venvs = pipx_venvs + uv_venvs

    return venvs


def print_installed_tools(tree: bool, tool: ToolType) -> None:
    venvs = find_installed_tools(tool)
    pipx_tools = sorted([pkg for pkg in venvs if pkg.tool == "pipx"], key=lambda x: x.name)
    uv_tools = sorted([pkg for pkg in venvs if pkg.tool == "uv"], key=lambda x: x.name)

    if pipx_tools and tree:
        print_installed_tree(pipx_tools, "pipx")
    elif pipx_tools:
        print_installed_table(pipx_tools, "pipx")

    if uv_tools and tree:
        print_installed_tree(uv_tools, "uv")
    elif uv_tools:
        print_installed_table(uv_tools, "uv")


def print_outdated_tools(is_prerelease: bool, tree: bool, tool: ToolType) -> None:
    venvs = find_installed_tools(tool)
    if not venvs:
        return

    packages_to_update = find_most_recent_version_on_pypi(venvs=venvs, is_prerelease=is_prerelease)
    if not packages_to_update:
        print("✨ Installed packages are all up to date ✨")
        return

    pipx_tools = sorted([pkg for pkg in packages_to_update if pkg.tool == "pipx"], key=lambda x: x.name)
    uv_tools = sorted([pkg for pkg in packages_to_update if pkg.tool == "uv"], key=lambda x: x.name)

    if pipx_tools and tree:
        print_outdated_tree(pipx_tools, "pipx")
    elif pipx_tools:
        print_outdated_table(pipx_tools, "pipx")

    if uv_tools and tree:
        print_outdated_tree(uv_tools, "uv")
    elif uv_tools:
        print_outdated_table(uv_tools, "uv")



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from contextlib import ExitStack

from packaging.version import InvalidVersion
from packaging.version import Version
from pypi_simple import DistributionPackage
from pypi_simple import NoSuchProjectError
from pypi_simple import PyPISimple
from rich.progress import BarColumn
from rich.progress import Progress
from rich.progress import TaskProgressColumn
from rich.progress import TextColumn
from rich.progress import TimeRemainingColumn

from piplexed.venvs import PackageInfo


def pypi_package_info(client: PyPISimple, package_name: str) -> list[DistributionPackage]:
    """List of installable packages from PyPI for a given project"""
    project_page = client.get_project_page(package_name)
    return project_page.packages


def latest_pypi_version(pkg_versions: list[DistributionPackage], is_prerelease: bool) -> Version:  # noqa: FBT001
    """Finds most recent package version that has an sdist distribution"""
    versions: list[Version] = []
    for pkg in pkg_versions:
        if pkg.version is not None and pkg.package_type == "sdist" and not pkg.is_yanked:
            try:
                pkg_vsn = Version(pkg.version)
            except InvalidVersion:
                continue
            if not pkg_vsn.is_prerelease or is_prerelease:
                versions.append(pkg_vsn)

    # use max(Version) instead of reversing order of package_page as recommended in PEP 700
    # https://peps.python.org/pep-0700/
    return max(versions)


def get_pypi_versions(client: PyPISimple, package: PackageInfo, is_prerelease: bool) -> PackageInfo:  # noqa: FBT001
    """Update a PackageInfo dataclass with the most recent version available on PyPI"""
    pypi_versions: list[DistributionPackage] = pypi_package_info(client=client, package_name=package.name)

    latest_version = latest_pypi_version(pypi_versions, is_prerelease=is_prerelease)

    package.latest_pypi_version = latest_version

    return package


def find_most_recent_version_on_pypi(*, venvs: list[PackageInfo], is_prerelease: bool) -> list[PackageInfo]:
    """Get most recent version on PyPI for a given list of packages (PackageInfo dataclasses)"""
    with ExitStack() as stack:
        client = stack.enter_context(PyPISimple())
        progress_bar = Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TimeRemainingColumn(),
            transient=True,
        )

        stack.enter_context(progress_bar)
        task = progress_bar.add_task("[red]Getting PyPI version data", total=len(venvs))

        executor = stack.enter_context(ThreadPoolExecutor(max_workers=5))

        results = [executor.submit(get_pypi_versions, client, pkg, is_prerelease) for pkg in venvs]

        updates = []
        for future in as_completed(results):
            try:
                result = future.result()
            except NoSuchProjectError:
                continue

            if result.newer_pypi_version():
                updates.append(result)
            progress_bar.update(task, advance=1)

    return sorted(updates, key=lambda x: x.name)



# SPDX-FileCopyrightText: 2023-present Andrew White <white-aj@outlook.com>
#
# SPDX-License-Identifier: MIT
from .app import print_installed_tools
from .app import print_outdated_tools
from .version import VERSION

__all__ = [
    "print_installed_tools",
    "print_outdated_tools",
    "VERSION",
]

__version__ = VERSION



from unittest.mock import patch

import pytest

from piplexed.utils import future_deprecation_warning


def test_future_deprecation_warning_with_replacement():
    with patch("warnings.warn") as mock_warn:
        future_deprecation_warning(
            reason="Functionality is being deprecated",
            replacement="Use the new_function() instead",
            deprecation_version="2.0.0",
        )
        mock_warn.assert_called_once_with(
            "Functionality is being deprecated. This will happen in 2.0.0. Use the new_function() instead",
            stacklevel=2,
            category=FutureWarning,
        )


def test_future_deprecation_warning_without_replacement():
    with patch("warnings.warn") as mock_warn:
        future_deprecation_warning(
            reason="Functionality is being deprecated", replacement=None, deprecation_version="2.0.0"
        )
        mock_warn.assert_called_once_with(
            "Functionality is being deprecated. This will happen in 2.0.0.", stacklevel=2, category=FutureWarning
        )


def test_future_deprecation_warning_empty_replacement():
    with patch("warnings.warn") as mock_warn:
        future_deprecation_warning(
            reason="Functionality is being deprecated", replacement="", deprecation_version="2.0.0"
        )
        mock_warn.assert_called_once_with(
            "Functionality is being deprecated. This will happen in 2.0.0.", stacklevel=2, category=FutureWarning
        )


def test_future_deprecation_warning_with_non_string_replacement():
    with pytest.raises(TypeError):
        future_deprecation_warning(
            reason="Functionality is being deprecated", replacement=1234, deprecation_version="2.0.0"
        )



from _pytest.capture import CaptureFixture
from packaging.version import Version

from piplexed.display.print_table import print_installed_table
from piplexed.display.print_table import print_outdated_table
from piplexed.venvs.pipx_venvs import PackageInfo


def test_print_installed_table(capsys: CaptureFixture[str]):
    packages = [
        PackageInfo("A", Version("1.0.1"), python="3.11.3"),
        PackageInfo("B", Version("2.2.1"), python="3.11.3"),
    ]
    print_installed_table(packages, "pipx")

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    headers = (
        "Package Name",
        "Installed Version",
        "Python Version",
    )
    assert output[0].startswith("PIPX Packages")
    assert all(header in output[2] for header in headers)
    assert output[4] == "│            A │             1.0.1 │         3.11.3 │"
    assert output[5] == "│            B │             2.2.1 │         3.11.3 │"
    assert output[6] == "└──────────────┴───────────────────┴────────────────┘"


def test_print_list_outdated(capsys: CaptureFixture[str]):
    packages = [
        PackageInfo("B", Version("1.5.0"), python="3.11.3", latest_pypi_version=Version("1.6.0")),
        PackageInfo("C", Version("2.6.0"), python="3.11.3", latest_pypi_version=Version("3.0a.0")),
    ]

    print_outdated_table(packages, "pipx")

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    headers = (
        "Package Name",
        "Installed Version",
        "PyPI Version",
    )

    assert output[0] == "PIPX Outdated Packages"
    assert all(header in output[2] for header in headers)
    assert output[4] == "│            B │             1.5.0 │        1.6.0 │"
    assert output[5] == "│            C │             2.6.0 │      3.0a0 ⚠ │"
    assert output[6] == "└──────────────┴───────────────────┴──────────────┘"



from __future__ import annotations

import json
import platform
import warnings
from pathlib import Path

from packaging.utils import canonicalize_name
from packaging.version import Version
from platformdirs import user_data_path

from piplexed.venvs import PackageInfo

PIPX_METADATA_VERSIONS = ["0.1", "0.2", "0.3", "0.4", "0.5"]
OS_PLATFORM = platform.system()


def pipx_home_paths_for_os(platform_: str) -> tuple[Path, list[Path]]:
    if platform_ == "Linux":
        default_pipx_home = Path(user_data_path("pipx"))
        fallback_pipx_homes = [Path.home() / ".local/pipx"]
    elif platform_ == "Windows":
        default_pipx_home = Path.home() / "pipx"
        fallback_pipx_homes = [Path.home() / ".local/pipx", Path(user_data_path("pipx"))]
    else:
        default_pipx_home = Path.home() / ".local/pipx"
        fallback_pipx_homes = [Path(user_data_path("pipx"))]

    return (default_pipx_home, fallback_pipx_homes)


DEFAULT_PIPX_HOME, FALLBACK_PIPX_HOMES = pipx_home_paths_for_os(OS_PLATFORM)


def get_local_venv() -> Path | None:
    if DEFAULT_PIPX_HOME.exists():
        return DEFAULT_PIPX_HOME / "venvs"

    for fallback_dir in FALLBACK_PIPX_HOMES:
        if fallback_dir.exists():
            return fallback_dir / "venvs"

    return None


PIPX_LOCAL_VENVS: Path | None = get_local_venv()


def is_metadata_version_valid(metadata_version: str, pipx_metadata_vsn: list[str] = PIPX_METADATA_VERSIONS) -> bool:
    return metadata_version in pipx_metadata_vsn


def installed_pipx_tools(venv_dir: Path | None = PIPX_LOCAL_VENVS) -> list[PackageInfo]:
    venvs: list[PackageInfo] = []
    if venv_dir is None or not venv_dir.exists():
        return venvs
    for env in venv_dir.iterdir():
        for item in env.iterdir():
            if item.name == "pipx_metadata.json":  # pragma: no branch
                with open(item) as f:
                    data = json.load(f)
                    metadata_version = data["pipx_metadata_version"]
                    if not is_metadata_version_valid(metadata_version):
                        warnings.warn(
                            f"{metadata_version} is an unknown (and untested) pipx metadata version,"
                            "results may be inaccurate",
                            stacklevel=2,
                            category=UserWarning,
                        )
                    pkg_data = PackageInfo(
                        name=canonicalize_name(data["main_package"]["package"]),
                        version=Version(data["main_package"]["package_version"]),
                        python=data["python_version"].split()[-1],
                    )
                    venvs.append(pkg_data)

    return venvs



from __future__ import annotations

from collections.abc import Iterable

from rich import print as rich_print
from rich.table import Table
from rich.text import Text

from piplexed.venvs import PackageInfo


def print_installed_table(packages: Iterable[PackageInfo], tool: str) -> None:
    table = Table(title=f"{tool.upper()} Packages")
    table.add_column("Package Name", justify="right", style="dark_orange", no_wrap=True)
    table.add_column("Installed Version", justify="right", style="deep_sky_blue1", no_wrap=True)
    table.add_column("Python Version", justify="right", style="green4", no_wrap=True)

    for pkg in packages:
        table.add_row(f"{pkg.name}", f"{pkg.version}", f"{pkg.python}")

    rich_print(table)


def print_outdated_table(package_data: Iterable[PackageInfo], tool: str) -> None:
    table = Table(title=f"{tool.upper()} Outdated Packages")
    table.add_column("Package Name", justify="right", style="dark_orange", no_wrap=True)
    table.add_column("Installed Version", justify="right", style="deep_sky_blue1", no_wrap=True)
    table.add_column("PyPI Version", justify="right", style="red3", no_wrap=True)

    for pkg in package_data:
        pypi_info = Text(f"{pkg.latest_pypi_version}", "green1")
        if pkg.latest_pypi_version is not None and pkg.latest_pypi_version.is_prerelease:
            pypi_info.append(" ⚠", "bright_yellow")

        table.add_row(pkg.name, f"{pkg.version}", pypi_info)

    rich_print(table)



from unittest.mock import patch

import pytest
from packaging.version import Version

from piplexed import app
from piplexed.venvs import PackageInfo
from piplexed.venvs import ToolType


@pytest.fixture()
def pipx_package():
    return PackageInfo(
        name="package1", version=Version("1.0.1"), python="3.12.7", latest_pypi_version=Version("1.5.0"), tool="pipx"
    )


@pytest.fixture()
def uv_package():
    return PackageInfo(
        name="package2", version=Version("2.0.5"), python="3.12.7", latest_pypi_version=Version("2.0.6"), tool="uv"
    )


@patch("piplexed.app.installed_pipx_tools")
def test_find_installed_tools_pipx(pipx_tools, pipx_package):
    expected = [pipx_package]
    pipx_tools.return_value = expected

    assert app.find_installed_tools(ToolType.PIPX) == expected


@patch("piplexed.app.installed_uv_tools")
def test_find_installed_tools_uv(uv_tools, uv_package):
    expected = [uv_package]
    uv_tools.return_value = expected

    assert app.find_installed_tools(ToolType.UV) == expected


@patch("piplexed.app.installed_uv_tools")
@patch("piplexed.app.installed_pipx_tools")
def test_find_installed_tools_all(pipx_tools, uv_tools, pipx_package, uv_package):
    expected = [
        pipx_package,
        uv_package,
    ]
    pipx_tools.return_value = [expected[0]]
    uv_tools.return_value = [expected[1]]

    assert app.find_installed_tools(ToolType.ALL) == expected


@patch("piplexed.app.PIPX_LOCAL_VENVS", None)
@patch("piplexed.app.installed_pipx_tools")
def test_find_installed_tools_pipx_none(pipx_tools, capsys):
    #  installed tools will return empty list if can't find any tools
    pipx_tools.return_value = []

    app.find_installed_tools(ToolType.PIPX)

    out, err = capsys.readouterr()
    assert not out
    error_output = [s.strip() for s in err.splitlines()]
    assert "Error" in error_output[0]


@patch("piplexed.app.PIPX_LOCAL_VENVS", "mock/pipx/venvs")
@patch("piplexed.app.installed_pipx_tools")
def test_find_installed_tools_pipx_no_tools_found(pipx_tools, capsys):
    #  installed tools will return empty list if can't find any tools
    pipx_tools.return_value = []

    app.find_installed_tools(ToolType.PIPX)

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    assert "Info" in output[0]
    assert "mock/pipx/venvs" in output[1]


@patch("piplexed.app.HAS_UV", False)
@patch("piplexed.app.installed_uv_tools")
def test_find_installed_tools_uv_none(uv_tools, capsys):
    uv_tools.return_value = []

    app.find_installed_tools(ToolType.UV)

    out, err = capsys.readouterr()
    assert not out
    error_output = [s.strip() for s in err.splitlines()]
    assert "Error" in error_output[0]
    assert "path" in error_output[2]


@patch("piplexed.app.UV_TOOL_DIR", None)
@patch("piplexed.app.HAS_UV", True)
@patch("piplexed.app.installed_uv_tools")
def test_find_installed_tools_uv_installed_no_tool_dir(uv_tools, capsys):
    uv_tools.return_value = []

    app.find_installed_tools(ToolType.UV)

    out, err = capsys.readouterr()
    assert not out
    error_output = [s.strip() for s in err.splitlines()]
    assert "Error" in error_output[0]
    assert "uv tool directory" in error_output[2]


@patch("piplexed.app.installed_uv_tools")
@patch("piplexed.app.installed_pipx_tools")
def test_find_installed_tools_all_but_uv_none(pipx_tools, uv_tools, pipx_package):
    expected = [
        pipx_package,
    ]
    pipx_tools.return_value = expected
    uv_tools.return_value = []

    assert app.find_installed_tools(ToolType.ALL) == expected


@patch("piplexed.app.installed_uv_tools")
@patch("piplexed.app.installed_pipx_tools")
def test_find_installed_tools_all_but_pipx_none(pipx_tools, uv_tools, uv_package):
    expected = [
        uv_package,
    ]
    pipx_tools.return_value = []
    uv_tools.return_value = expected

    assert app.find_installed_tools(ToolType.ALL) == expected


@patch("piplexed.app.installed_uv_tools")
@patch("piplexed.app.installed_pipx_tools")
def test_find_installed_tools_all_but_pipx_and_uv_none(pipx_tools, uv_tools):
    pipx_tools.return_value = []
    uv_tools.return_value = []

    result = app.find_installed_tools(ToolType.ALL)
    assert result == []


@patch("piplexed.app.find_installed_tools")
def test_print_installed_tools_pipx_table(mock_tools, pipx_package, capsys: pytest.CaptureFixture[str]):
    mock_tools.return_value = [pipx_package]
    app.print_installed_tools(tree=False, tool=ToolType.PIPX)

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    headers = (
        "Package Name",
        "Installed Version",
        "Python Version",
    )
    assert output[0].startswith("PIPX Packages")
    assert all(header in output[2] for header in headers)
    assert output[4] == "│     package1 │             1.0.1 │         3.12.7 │"


@patch("piplexed.app.find_installed_tools")
def test_print_installed_tools_uv_table(mock_tools, uv_package, capsys: pytest.CaptureFixture[str]):
    mock_tools.return_value = [uv_package]
    app.print_installed_tools(tree=False, tool=ToolType.UV)

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    headers = (
        "Package Name",
        "Installed Version",
        "Python Version",
    )
    assert output[0].startswith("UV Packages")
    assert all(header in output[2] for header in headers)
    assert output[4] == "│     package2 │             2.0.5 │         3.12.7 │"


@patch("piplexed.app.find_installed_tools")
def test_print_installed_tools_all_table(mock_tools, pipx_package, uv_package, capsys: pytest.CaptureFixture[str]):
    mock_tools.return_value = [pipx_package, uv_package]
    app.print_installed_tools(tree=False, tool=ToolType.ALL)

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    headers = (
        "Package Name",
        "Installed Version",
        "Python Version",
    )
    assert output[0].startswith("PIPX Packages")
    assert all(header in output[2] for header in headers)
    assert output[4] == "│     package1 │             1.0.1 │         3.12.7 │"
    assert output[6].startswith("UV Packages")
    assert all(header in output[8] for header in headers)
    assert output[10] == "│     package2 │             2.0.5 │         3.12.7 │"


@patch("piplexed.app.find_installed_tools")
def test_print_installed_tools_pipx_tree(mock_tools, pipx_package, capsys: pytest.CaptureFixture[str]):
    mock_tools.return_value = [pipx_package]
    app.print_installed_tools(tree=True, tool=ToolType.PIPX)

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    assert output[0].startswith("PIPX Packages")
    assert output[1] == "└── Package1"
    assert output[2] == "├── version - 1.0.1"
    assert output[3] == "└── python - 3.12.7"


@patch("piplexed.app.find_installed_tools")
def test_print_installed_tools_uv_tree(mock_tools, uv_package, capsys: pytest.CaptureFixture[str]):
    mock_tools.return_value = [uv_package]
    app.print_installed_tools(tree=True, tool=ToolType.UV)

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    assert output[0].startswith("UV Packages")
    assert output[1] == "└── Package2"
    assert output[2] == "├── version - 2.0.5"
    assert output[3] == "└── python - 3.12.7"


@patch("piplexed.app.find_installed_tools")
def test_print_installed_tools_all_tree(mock_tools, pipx_package, uv_package, capsys: pytest.CaptureFixture[str]):
    mock_tools.return_value = [pipx_package, uv_package]
    app.print_installed_tools(tree=True, tool=ToolType.ALL)

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    assert output[0].startswith("PIPX Packages")
    assert output[1] == "└── Package1"
    assert output[2] == "├── version - 1.0.1"
    assert output[3] == "└── python - 3.12.7"
    assert output[4].startswith("UV Packages")
    assert output[5] == "└── Package2"
    assert output[6] == "├── version - 2.0.5"
    assert output[7] == "└── python - 3.12.7"


@patch("piplexed.app.find_most_recent_version_on_pypi")
@patch("piplexed.app.find_installed_tools")
def test_print_outdated_tools_pipx_table(mock_tools, mock_outdated, pipx_package, capsys):
    mock_tools.return_value = [pipx_package]
    mock_outdated.return_value = [pipx_package]
    app.print_outdated_tools(is_prerelease=False, tree=False, tool=ToolType.PIPX)

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    headers = (
        "Package Name",
        "Installed Version",
        "PyPI Version",
    )
    assert output[0].startswith("PIPX Outdated Packages")
    assert all(header in output[2] for header in headers)
    assert output[4] == "│     package1 │             1.0.1 │        1.5.0 │"


@patch("piplexed.app.find_most_recent_version_on_pypi")
@patch("piplexed.app.find_installed_tools")
def test_print_outdated_tools_uv_table(mock_tools, mock_outdated, uv_package, capsys):
    mock_tools.return_value = [uv_package]
    mock_outdated.return_value = [uv_package]
    app.print_outdated_tools(is_prerelease=False, tree=False, tool=ToolType.UV)

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    headers = (
        "Package Name",
        "Installed Version",
        "PyPI Version",
    )
    assert output[0].startswith("UV Outdated Packages")
    assert all(header in output[2] for header in headers)
    assert output[4] == "│     package2 │             2.0.5 │        2.0.6 │"


@patch("piplexed.app.find_most_recent_version_on_pypi")
@patch("piplexed.app.find_installed_tools")
def test_print_outdated_tools_all_table(mock_tools, mock_outdated, pipx_package, uv_package, capsys):
    mock_tools.return_value = [pipx_package, uv_package]
    mock_outdated.return_value = [pipx_package, uv_package]
    app.print_outdated_tools(is_prerelease=False, tree=False, tool=ToolType.ALL)

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    headers = (
        "Package Name",
        "Installed Version",
        "PyPI Version",
    )
    assert output[0].startswith("PIPX Outdated Packages")
    assert all(header in output[2] for header in headers)
    assert output[4] == "│     package1 │             1.0.1 │        1.5.0 │"
    assert output[6].startswith("UV Outdated Packages")
    assert all(header in output[8] for header in headers)
    assert output[10] == "│     package2 │             2.0.5 │        2.0.6 │"


@patch("piplexed.app.find_most_recent_version_on_pypi")
@patch("piplexed.app.find_installed_tools")
def test_print_outdateded_packages_pipx_tree(
    mock_tools, mock_outdated, pipx_package, capsys: pytest.CaptureFixture[str]
):
    mock_tools.return_value = [pipx_package]
    mock_outdated.return_value = [pipx_package]
    app.print_outdated_tools(is_prerelease=False, tree=True, tool=ToolType.PIPX)

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    assert output[0].startswith("PIPX Outdated Packages")
    assert output[1] == "└── Package1"
    assert output[2] == "├── installed version - 1.0.1"
    assert output[3] == "└── PyPI version - 1.5.0"


@patch("piplexed.app.find_most_recent_version_on_pypi")
@patch("piplexed.app.find_installed_tools")
def test_print_outdateded_packages_uv_tree(mock_tools, mock_outdated, uv_package, capsys: pytest.CaptureFixture[str]):
    mock_tools.return_value = [uv_package]
    mock_outdated.return_value = [uv_package]
    app.print_outdated_tools(is_prerelease=False, tree=True, tool=ToolType.UV)

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    assert output[0].startswith("UV Outdated Packages")
    assert output[1] == "└── Package2"
    assert output[2] == "├── installed version - 2.0.5"
    assert output[3] == "└── PyPI version - 2.0.6"


@patch("piplexed.app.find_most_recent_version_on_pypi")
@patch("piplexed.app.find_installed_tools")
def test_print_outdateded_packages_all_tree(
    mock_tools, mock_outdated, pipx_package, uv_package, capsys: pytest.CaptureFixture[str]
):
    mock_tools.return_value = [pipx_package, uv_package]
    mock_outdated.return_value = [pipx_package, uv_package]
    app.print_outdated_tools(is_prerelease=False, tree=True, tool=ToolType.ALL)

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    assert output[0].startswith("PIPX Outdated Packages")
    assert output[1] == "└── Package1"
    assert output[2] == "├── installed version - 1.0.1"
    assert output[3] == "└── PyPI version - 1.5.0"
    assert output[4].startswith("UV Outdated Packages")
    assert output[5] == "└── Package2"
    assert output[6] == "├── installed version - 2.0.5"
    assert output[7] == "└── PyPI version - 2.0.6"


@patch("piplexed.app.find_most_recent_version_on_pypi")
@patch("piplexed.app.find_installed_tools")
def test_print_outdateded_packages_pipx_no_updates(
    mock_tools, mock_outdated, pipx_package, capsys: pytest.CaptureFixture[str]
):
    mock_tools.return_value = [pipx_package]
    mock_outdated.return_value = []
    app.print_outdated_tools(is_prerelease=False, tree=True, tool=ToolType.PIPX)

    out, err = capsys.readouterr()
    assert not err
    assert out.strip() == "✨ Installed packages are all up to date ✨"


@patch("piplexed.app.find_most_recent_version_on_pypi")
@patch("piplexed.app.find_installed_tools")
def test_print_outdateded_packages_uv_no_updates(
    mock_tools, mock_outdated, uv_package, capsys: pytest.CaptureFixture[str]
):
    mock_tools.return_value = [uv_package]
    mock_outdated.return_value = []
    app.print_outdated_tools(is_prerelease=False, tree=True, tool=ToolType.UV)

    out, err = capsys.readouterr()
    assert not err
    assert out.strip() == "✨ Installed packages are all up to date ✨"


@patch("piplexed.app.find_most_recent_version_on_pypi")
@patch("piplexed.app.find_installed_tools")
def test_print_outdateded_packages_all_no_updates(
    mock_tools, mock_outdated, pipx_package, uv_package, capsys: pytest.CaptureFixture[str]
):
    mock_tools.return_value = [pipx_package, uv_package]
    mock_outdated.return_value = []
    app.print_outdated_tools(is_prerelease=False, tree=True, tool=ToolType.ALL)

    out, err = capsys.readouterr()
    assert not err
    assert out.strip() == "✨ Installed packages are all up to date ✨"



from pathlib import Path
from unittest.mock import patch

import pytest
from packaging.version import Version

from piplexed.venvs import PackageInfo
from piplexed.venvs.uv_venvs import find_uv
from piplexed.venvs.uv_venvs import find_uv_tool_dir
from piplexed.venvs.uv_venvs import get_python_path
from piplexed.venvs.uv_venvs import installed_uv_tools
from piplexed.venvs.uv_venvs import uv_tool_version


@patch("piplexed.venvs.uv_venvs.shutil.which")
def test_find_uv_when_installed(mock_uv_path):
    mock_uv_path.return_value = "mock/path/uv"

    result = find_uv()
    assert result == (True, "mock/path/uv")


@patch("piplexed.venvs.uv_venvs.shutil.which")
def test_find_uv_not_installed(mock_uv_path):
    mock_uv_path.return_value = None

    result = find_uv()
    assert result == (False, "uv")


def test_find_uv_tool_dir_present(tmp_path, fp):
    mock_uv_path = tmp_path / "uv"
    mock_tool_dir = mock_uv_path / "tool"
    mock_tool_dir.mkdir(parents=True)
    fp.register([str(mock_uv_path), "tool", "dir"], stdout=str(mock_tool_dir))

    assert find_uv_tool_dir(str(mock_uv_path)) == str(mock_tool_dir)


def test_find_uv_tool_dir_not_present(tmp_path, fp):
    mock_uv_path = tmp_path / "uv"
    mock_tool_dir = mock_uv_path / "tool"
    fp.register([str(mock_uv_path), "tool", "dir"], stdout=str(mock_tool_dir))

    assert find_uv_tool_dir(str(mock_uv_path)) is None


def test_uv_tool_dir_failed_returncode(fp):
    def callback_function(process):
        process.returncode = 1

    fp.register(["mock/path/uv", "tool", "dir"], stdout="uv/tool/dir", callback=callback_function)

    assert find_uv_tool_dir("mock/path/uv") is None


def test_uv_tool_version_windows(tmp_path, fp):
    mock_tool_dir = tmp_path / "package1"
    mock_tool_python = mock_tool_dir / "Scripts" / "python.exe"

    script = "from importlib.metadata import version; print(version('package1'))"
    fp.register([mock_tool_python, "-c", script], stdout="1.0.1")

    result = uv_tool_version(mock_tool_dir, "Windows")

    assert result == "1.0.1"


def test_uv_tool_version_linux(tmp_path, fp):
    mock_tool_dir = tmp_path / "package1"
    mock_tool_python = mock_tool_dir / "bin" / "python"

    script = "from importlib.metadata import version; print(version('package1'))"
    fp.register([mock_tool_python, "-c", script], stdout="1.0.1")

    result = uv_tool_version(mock_tool_dir, "Linux")

    assert result == "1.0.1"


@pytest.mark.parametrize(
    "op_system, expected",
    [
        pytest.param("Windows", Path("Scripts") / "python.exe", id="windows"),
        pytest.param("Linux", Path("bin") / "python", id="linux"),
        pytest.param("Darwin", Path("bin") / "python", id="macOS"),
    ],
)
def test_get_python_path(tmp_path, op_system, expected):
    mock_venv = tmp_path / "package1"
    result = get_python_path(mock_venv, op_system)
    assert result == Path.joinpath(mock_venv, expected)


@patch("piplexed.venvs.uv_venvs.get_python_path")
def test_installed_uv_tools_windows(mock_py_path, tmp_path, fp):
    mock_tools_venv = tmp_path / "venvs"
    mock_tools_venv.mkdir()

    test_ignore_path = mock_tools_venv / ".tester"
    test_ignore_path.mkdir()

    package_1 = mock_tools_venv / "package1"
    package_1.mkdir()

    mock_py_path.return_value = package_1 / "Scripts" / "python.exe"

    package_version_cmd = f"from importlib.metadata import version; print(version('{package_1.name}'))"
    python_version_cmd = "import sys; print(sys.version.split()[0])"
    fp.register([package_1 / "Scripts" / "python.exe", "-c", package_version_cmd], stdout="1.0.1")
    fp.register([package_1 / "Scripts" / "python.exe", "-c", python_version_cmd], stdout="3.12.7")

    result = installed_uv_tools(mock_tools_venv, "Windows")
    assert result == [PackageInfo(name="package1", version=Version("1.0.1"), python="3.12.7", tool="uv")]


@patch("piplexed.venvs.uv_venvs.get_python_path")
def test_installed_uv_tools_linux(mock_py_path, tmp_path, fp):
    mock_tools_venv = tmp_path / "venvs"
    mock_tools_venv.mkdir()

    test_ignore_path = mock_tools_venv / ".tester"
    test_ignore_path.mkdir()

    package_1 = mock_tools_venv / "package1"
    package_1.mkdir()

    mock_py_path.return_value = package_1 / "bin" / "python"

    package_version_cmd = f"from importlib.metadata import version; print(version('{package_1.name}'))"
    python_version_cmd = "import sys; print(sys.version.split()[0])"
    fp.register([package_1 / "bin" / "python", "-c", package_version_cmd], stdout="1.0.1")
    fp.register([package_1 / "bin" / "python", "-c", python_version_cmd], stdout="3.12.7")

    result = installed_uv_tools(mock_tools_venv, "Linux")
    assert result == [PackageInfo(name="package1", version=Version("1.0.1"), python="3.12.7", tool="uv")]
